//
//  EDAMClient.h
//  Evernote
//
//  Created by Andrew McGeachie on 9/18/07.
//  Copyright 2007 Evernote Corp. All rights reserved.
//

#ifndef EDAMClient_HEADER_GUARD
#define EDAMClient_HEADER_GUARD

#import <EDAM/EDAMCatalog.h>
#import <EDAM/EDAMCommerce.h>
#import <EDAM/EDAMNoteStore.h>
#import <EDAM/EDAMUserStore.h>
#import <EDAM/EDAMUtility.h>


/**
 * Maintains a UserStore and a NoteStore and handles work of maintaining a valid authentication token.  The authToken property can be used to fetch a valid authentication token.  When this property is first read it automatically retrieves a new authentication token from the server.  On subsequent reads the authentication token is automagically either refreshed or replaced.
 */
@protocol EDAMClient <NSObject, NSCopying>

@property(retain, nonatomic) NSString * username;
@property(retain, nonatomic) NSString * password;

@property(readonly,retain, nonatomic) NSString * authToken;


/**
 * Make sure that our version of the EDAM client is compatible with the 
 * current server version.
 *
 * https://bugz.evernote.com/fogbugz/default.asp?164241
 * The result of this call may be cached until the app 
 * quits.
 */
- (BOOL) checkVersion;


/**
 * Return YES if we have a valid authentication token from the server.
 */
- (BOOL) hasValidAuthenticationToken;


/**
 * Make sure we are authenticated with the server.
 * @return The authentication info from the server.
 */
- (EDAMAuthenticationResult *) authenticate;


/**
 * Submits a block for asynchronous execution on the user store queue and returns immediately.
 */
- (void) performUserStoreBlock:(void (^)(id<EDAMUserStore>))block;


/**
 * Submits a block object for execution on the user store queue and waits until that block completes.
 */
- (void) performUserStoreBlockAndWait:(void (^)(id<EDAMUserStore>))block;


/**
 * Submits a block for asynchronous execution on the note store queue and returns immediately.
 */
- (void) performNoteStoreBlock:(void (^)(id<EDAMNoteStore>))block;


/**
 * Submits a block object for execution on the note store queue and waits until that block completes.
 */
- (void) performNoteStoreBlockAndWait:(void (^)(id<EDAMNoteStore>))block;

/**
 * Submits a block object for execution on the note store queue and waits until that block completes (in a background session).
 */
- (void) performNoteStoreBlockInSessionAndWait:(void (^)(id<EDAMNoteStore>))block WithSessionIdentifier:(NSString*)sessionId;


/**
 * Submits a block for asynchronous execution on the note store queue and returns immediately.
 */
- (void) performCatalogBlock:(void (^)(id<EDAMCatalog>))block;


/**
 * Submits a block object for execution on the note store queue and waits until that block completes.
 */
- (void) performCatalogBlockAndWait:(void (^)(id<EDAMCatalog>))block;


/**
 * Submits a block for asynchronous execution on the note store queue and returns immediately.
 */
- (void) performCommerceBlock:(void (^)(id<EDAMCommerce>))block;


/**
 * Submits a block object for execution on the note store queue and waits until that block completes.
 */
- (void) performCommerceBlockAndWait:(void (^)(id<EDAMCommerce>))block;


/**
 * Submits a block for asynchronous execution on the note store queue and returns immediately.
 */
- (void) performUtilityBlock:(void (^)(id<EDAMUtility>))block;


/**
 * Submits a block object for execution on the note store queue and waits until that block completes.
 */
- (void) performUtilityBlockAndWait:(void (^)(id<EDAMUtility>))block;


#if EDAMCLIENT_INCLUDE_MESSAGE_CLIENT
/**
 * Submits a block for asynchronous execution on the message store queue and returns immediately.
 */
- (void) performMessageStoreBlock:(void (^)(id<EDAMMessageStore>))block;


/**
 * Submits a block for execution on the message store queue and waits until that block completes.
 */
- (void) performMessageStoreBlockAndWait:(void (^)(id<EDAMMessageStore>))block;
#endif

@end

#endif