//
//  ENBootstrapProfileManager.h
//  EverNote
//
//  Created by Brandon Volbright on 3/7/12.
//  Copyright (c) 2012 Evernote. All rights reserved.
//


// GEECH - 11/19/13
// This is the bootstrap profile manager from the mac client.  at some point it should be merged into the
// bootstrap manager used by other clients.  but not right now.


#import <Foundation/Foundation.h>

@class EDAMBootstrapProfile;


typedef enum {
  ENBPMStatus_None = 0,     // not yet started
  ENBPMStatus_Failed,       // profile retrieval failed
  ENBPMStatus_Working,      // currently retrieving profiles
  ENBPMStatus_Done,         // successfully retrieved profiles
} ENBPMStatus;


// sent when the currentProfile property changes
extern NSString *const ENBootstrapCurrentProfileChangedNotification;

// sent when fresh profile data is received from a bootstrap server
extern NSString *const ENBootstrapProfileDataReceivedNotification;

// sent when retrieval of profile data fails
extern NSString *const ENBootstrapProfileRetrievalFailedNotification;

// name of the international profile
extern NSString *const ENBootstrapInternationalProfileName;
extern NSString *const ENBootstrapChinaProfileName;

extern NSString *const kENSubstituteServerName; // a substitute International server name via the defaults


/**
 * Singleton for retrieving and managing Evernote "bootstrap" profiles.
 */
@interface ENBootstrapManager : NSObject {
  ENBPMStatus mStatus;
  NSArray *mProfiles;
  EDAMBootstrapProfile *mPreferredProfile;
  EDAMBootstrapProfile *mCurrentProfile;
  EDAMBootstrapProfile *mPlaceholderProfileInternational;
  EDAMBootstrapProfile *mPlaceholderProfileChina;
  BOOL mHasLoadedData;
}

/** Whether the manager is working, failed, done, etc. */
@property (readonly,nonatomic) ENBPMStatus status;

/** Profiles received from the bootstrap service. */
@property (readonly,nonatomic) NSArray *profiles;

/** Preferred profile, as indicated by the bootstrap service. */
@property (readonly,nonatomic) EDAMBootstrapProfile *preferredProfile;

/** Currently selected profile. */
@property (retain,nonatomic) EDAMBootstrapProfile *currentProfile;

/** Whether the manager has profile data that is not placeholder data. */
@property (readonly,nonatomic) BOOL hasNonPlaceholderData;


/**
 * Retrieve the singleton instance.
 */
+ (ENBootstrapManager *)sharedManager;


/**
 * Convenience class-level method to get current profile.
 */
+ (EDAMBootstrapProfile *)currentBootstrapProfile;


/**
 * Retrieve bootstrap info.
 * If cached bootstrap info is available, does not attempt to contact a bootstrap server.
 * Otherwise, starts retrieving bootstrap info from a bootstrap server asynchronously.
 */
- (void)retrieveBootstrapInfo;


/**
 * Retrieve bootstrap info.
 * @param force If YES, contact a bootstrap server to retrieve data. If NO, contact a bootstrap
 * server only if there is no cached bootstrap info available.
 * Bootstrap server communication is asynchronous.
 */
- (void)retrieveBootstrapInfo:(BOOL)force;


/**
 * Convenience methods for retrieving feature enabled states.
 * Equivalent to [[[[ENBootstrapProfileManager sharedInstance] currentProfile] settings] enableXXX].
 * These methods are the preferred way to check whether a feature is enabled or not.
 */
+ (BOOL)facebookSharingEnabled;
+ (BOOL)giftSubscriptionsEnabled;
+ (BOOL)supportTicketsEnabled;
+ (BOOL)sharedNotebooksEnabled;
+ (BOOL)singleNoteSharingEnabled;
+ (BOOL)sponsoredAccountsEnabled;
+ (BOOL)twitterSharingEnabled;
+ (BOOL)linkedInSharingEnabled;


/**
 * Convenience method for access to the serviceHost property of the current profile.
 * Equivalent to [[[[ENBootstrapProfileManager sharedInstance] currentProfile] settings] serviceHost].
 * Returns nil if there is no current profile (which should be considered an exceptional state).
 */
+ (NSString *)edamServiceHost;


/**
 * Convenience method to list all possible EDAM service host names.
 */
+ (NSArray *)allPossibleEDAMServiceHosts;


/**
 * Convenience method for access to the marketingUrl property of the current profile.
 * Equivalent to [[[[ENBootstrapProfileManager sharedInstance] currentProfile] settings] marketingUrl].
 * Returns nil if there is no current profile (which should be considered an exceptional state).
 */
+ (NSString *)marketingURL;


/**
 * Convenience method for access to the supportUrl property of the current profile.
 * Equivalent to [[[[ENBootstrapProfileManager sharedInstance] currentProfile] settings] supportUrl].
 * Returns nil if there is no current profile (which should be considered an exceptional state).
 */
+ (NSString *)supportURL;


/**
 * Convenienve method for access to the accountEmailDomain property of the current profile.
 * Equivalent to [[[[ENBootstrapProfileManager sharedInstance] currentProfile] settings] accountEmailDomain].
 * Returns nil if there is no current profile (which should be considered an exceptional state).
 */
+ (NSString *)accountEmailDomain;

/**
 * Convenienve method for access to the cardscanURL property of the current profile.
 * Equivalent to [[[[ENBootstrapProfileManager sharedInstance] currentProfile] settings] cardscanUrl].
 * Returns nil if there is no current profile (which should be considered an exceptional state).
 */
+ (NSString *)cardscanURL;

/**
 * Find the bootstrap profile (if any) corresponding to the given profile name.
 */
+ (EDAMBootstrapProfile *)profileForName:(NSString *)name;


/**
 * Replace the current profile with the other profile in the current profile set.
 * If there are fewer than two profiles in the current profile set, no action is taken.
 */
- (void)toggleCurrentProfile;


/**
 * Whether the current profile is the International profile.
 */
+ (BOOL)isInternationalProfileCurrent;


/**
 * Whether the current profile is the China profile
 */
+ (BOOL) isChinaProfileCurrent;


/**
 * Start the bootstrap manager.  Will start fetching a profile if we don't have one
 */
+ (void) start;


@end
