//
//  ENSpotlightUtils.h
//  EvernoteSpotlightImporter
//
//  Created by Benjamin Chan on 7/11/14.
//  Copyright (c) 2014 Evernote Corporation. All rights reserved.
//

#import <Foundation/Foundation.h>
#import <CoreData/CoreData.h>
#import <CoreNote/CoreNote.h>

/**
 * A Helper object through which Spotlight stores should be accessed.
 * =====================================
 * DEPENDENT ON CORENOTE. Do not import this header if CoreNote access is lacking.
 */
@interface ENSpotlightUtils : NSObject

/**
 * Returns a dictionary containing information needed
 * to instantiate the relevant persistent store from an external record.
 * =====================================
 * A custom version of NSPersistentStoreCoordinator's -elementsDerivedFromExternalRecordURL:, for use with
 * our custom-generated external records. Note that as of 7/2014, the Apple version works just fine on our custom records,
 * but we provide our own implementation for robustness. See the Apple documentation for what this method attempts to mirror.
 * Provides values for:
 *      NSEntityNameInPathKey   (string name of the entity)
 *      NSStoreUUIDInPathKey    (string, persistent store's identifier)
 *      NSStorePathKey          (string, path to the actual persistent store file)
 *      NSModelPathKey          (string, path to the model file to use to load the persistent store)
 *      NSObjectURIKey          (NSURL with identifier of the specific managed object)
 */
+ (NSDictionary *)storeAttributesDerivedFromENExternalRecordURL:(NSURL *)extURL;



#pragma mark Data Store Access

/**
 * Returns (NSManagedObject) the managed object
 * associated with the metadata file at the given path.
 * If specified, also passes back a dictionary with the results of
 * storeAttributesDerivedFrom... as defined above.
 * =======================================
 * Returns (nil) if not successful.
 * Configures modelURL and storeURL.
 */
- (NSManagedObject *)managedObjectAtExternalRecordPath:(NSString *)extPath;
- (NSManagedObject *)managedObjectAtExternalRecordPath:(NSString *)extPath storeAttributes:(NSDictionary **)outReturnPathInfo;

/**
 * Same as -managedObjectAtExternalRecordPath, but also checks
 * if the MO is an ENNoteMO and returns nil if it is not.
 */
- (ENNoteMO *)noteManagedObjectAtExternalRecordPath:(NSString *)extPath;
- (ENNoteMO *)noteManagedObjectAtExternalRecordPath:(NSString *)extPath storeAttributes:(NSDictionary **)outReturnPathInfo;


#pragma mark Work on a Managed Object
/** ---------------------------------
 * To use these methods, first get a managed object from (e.g.) -managedObjectAtExternalRecordPath
 * and pathInfo if necessary, and then pass back to the method in question.
 */

/**
 * Returns (string) the backing directory for the note
 * to which the metadata file points.
 * ======================================
 * Uses hardcoded paths to shimmy through to the backing directory of a note.
 * --------------------------------------
 * This method is easily broken if the directory structure of a note's backing location
 * is changed. It is, however, more robust and easier than depending on CoreNote to tell us
 * where things are relative to a store/metadatafile.
 * Also note that when calling this method often, it is most efficient to group notes
 * stored in the same store, as to avoid having to switch stores all the time.
 */
- (NSString *)backingDirectoryForNote:(ENNoteMO *)note withStoreAttributes:(NSDictionary *)storeAttr;

#pragma mark Custom Data Store Access

// Set the following two properties to configure Data Store Access.
// Currently only works with SQLite Core Data stores.
@property (nonatomic, strong) NSURL *modelURL;
@property (nonatomic, strong) NSURL *storeURL;
// Use these methods to access managed objects from the most recently used data store
@property (readonly, strong, nonatomic) NSPersistentStoreCoordinator *persistentStoreCoordinator;
@property (readonly, strong, nonatomic) NSManagedObjectModel *managedObjectModel;
@property (readonly, strong, nonatomic) NSManagedObjectContext *managedObjectContext;



@end
