/* Message View -- a component used to render the conversation for Evernote
 * Messages. Written by Andrei Thorp, 2014 <athorp@evernote.com> */

/**
 * Container / namespace for the internal MVC structure of the program.
 *
 * @type {{
 *     Server: *,
 *     Runtime: *,
 *     Template: *,
 *     View: *,
 *     Controller: *,
 *     Helper: *,
 *     Defines: *
 * }}
 */
var Internal = {
    Server: null,
    Runtime: null,
    Template: null,
    View: null,
    Controller: null,
    Helper: null,
    Defines: null
};

/**
 * Container / namespace for the exports for the host environment.
 *
 * @type {*}
 */
var MessageView = {};

/**
 * @typedef {{
 *     id: string,
 *     type: string,
 *     title: string
 * }}
 */
Internal.Attachment;

/**
 * @typedef {{
 *     id: string,
 *     senderID: string,
 *     senderName: string,
 *     senderImage: string,
 *     body: string,
 *     sentAt: string,
 *     attachments: Array.<Internal.Attachment>
 * }}
 */
Internal.Message;

Internal.Defines = function() {
    var me = {};

    /**
     * How much a message must be on screen for it to be considered 'visible'
     *
     * @type {number} percent
     */
    me.MESSAGE_ON_SCREEN_PERCENT = 55;

    /**
     * How long we should wait to decide that the user has stopped scrolling
     *
     * @type {number} milliseconds
     */
    me.SCROLL_JITTER_TIMEOUT = 300;

    /**
     * List of which elements and attributes we support when rendering a message
     * body to the screen.
     *
     * @type {*}
     */
    me.STYLE_WHITELIST = {
        'a': ['href'],
        'br': []
    };

    /** @type {number} */ me.NODE_ELEMENT_NODE = 1;
    /** @type {number} */ me.NODE_ATTRIBUTE_NODE = 2;
    /** @type {number} */ me.NODE_TEXT_NODE = 3;
    /** @type {number} */ me.NODE_CDATA_SECTION_NODE = 4;
    /** @type {number} */ me.NODE_ENTITY_REFERENCE_NODE = 5;
    /** @type {number} */ me.NODE_ENTITY_NODE = 6;
    /** @type {number} */ me.NODE_PROCESSING_INSTRUCTION_NODE = 7;
    /** @type {number} */ me.NODE_COMMENT_NODE = 8;
    /** @type {number} */ me.NODE_DOCUMENT_NODE = 9;
    /** @type {number} */ me.NODE_DOCUMENT_TYPE_NODE = 10;
    /** @type {number} */ me.NODE_DOCUMENT_FRAGMENT_NODE = 11;
    /** @type {number} */ me.NODE_NOTATION_NODE = 12;

    return me;
}();

Internal.Server = function() {
    var me = {};

    /**
     * Opens the given attachment by id in the host environment.
     *
     * @param {string} id Attachment ID
     *
     * @suppress {undefinedVars|missingProperties}
     */
    me.openAttachment = function(id) {
        console.log('To Server: open attachment', id);

        try {
            Evernote.openAttachment(id);
        } catch (err) {
            console.warn('Couldn\'t run native function:', err.message);
        }
    };

    /**
     * Notifies the host environment that a message has been viewed.
     *
     * @param {string} id Message ID
     *
     * @suppress {undefinedVars|missingProperties}
     */
    me.messageViewed = function(id) {
        console.log('To Server: message viewed', id);

        try {
            Evernote.messageViewed(id);
        } catch (err) {
            console.warn('Couldn\'t run native function:', err.message);
        }
    };

    return me;
}();

Internal.Runtime = function() {
    var me = {};

    /**
     * Checks if the current platform is Mac.
     *
     * @return {boolean}
     */
    me.isMacPlatform = function() {
        if (navigator.platform.indexOf("Mac") >= 0) {
            return true;
        } else {
            return false;
        }
    };

    /**
     * Loads the CSS file for the specified dist target.
     *
     *
     * @param {string} opt_type either 'mac' or 'win'
     */
    me.loadCSS = function(opt_type) {
        /** @type {string} */
        var cssFile = '<link rel="stylesheet" href="css/win-main.css">';

        if (opt_type == null) {
            if (me.isMacPlatform()) {
                opt_type = 'mac';
            } else {
                opt_type = 'win';
            }
        }

        if (opt_type === 'mac') {
            cssFile = '<link rel="stylesheet" href="css/mac-main.css">';
        } else if (opt_type === 'win') {
            cssFile = '<link rel="stylesheet" href="css/win-main.css">';
        }

        /** @type {Node} */ var div = document.createElement('div');
        div.innerHTML = cssFile;
        document.body.appendChild(div.childNodes[0]);
    };

    /**
     * Attempts to load the given image's src based on the value of its
     * data-desired-src attribute. This happens by first trying to background
     * load the image in general, and if that works, then we set the src.
     * If it fails, we load the placeholder image instead.
     *
     * @param {Node} image
     */
    me.tryLoadImage = function(image) {
        /** @type {Image} */ var backgroundImage = new Image();

        backgroundImage.onload = function _imageOnLoad() {
            image.src = image.getAttribute('data-desired-src');
            backgroundImage.onload = null;
            backgroundImage.onerror = null;
        };

        backgroundImage.onerror = function _imageOnError() {
            image.src = Internal.Helper.getPlaceholderImageName();
            backgroundImage.onload = null;
            backgroundImage.onerror = null;
        };

        backgroundImage.src = image.getAttribute('data-desired-src');
    };

    /**
     * Periodically checks all of the images in the message list to see if
     * their URLs loaded successfully. If they haven't, it forces them to
     * retry. (See README.md for why we do this.)
     */
    me.imageLoadLoop = function() {
        /** @type {Node} */ var messageList;
        /** @type {NodeList} */ var images;
        /** @type {number} */ var timerID;

        timerID = setInterval(function _imageLoadLoop() {
            messageList = Internal.Helper.getMessageList();

            // Probably means we're running unit tests, but best to just bail
            // out if things get this crazy in general.
            if (messageList == null) {
                console.warn('Couldnt find messageList, exiting imageLoadLoop');
                clearInterval(timerID);
                return;
            }

            images = messageList.querySelectorAll('img');

            for (var i = images.length - 1; i >= 0; i--) {
                me.tryLoadImage(images[i]);
            }
        }, 5000);
    };

    /**
     * Looks for changes in the window's scroll position and intelligently
     * informs the native client when a new message is viewed.
     */
    me.messagesOnScreenWatcher = function() {
        /** @type {number} */ var timeoutID = -1;
        /** @type {Array.<Node>} */ var lastSeenMessages = [];
        /** @type {Array.<Node>} */ var currentMessages = [];
        /** @type {Node} */ var msg = null;

        var listener = function() {
            clearTimeout(timeoutID);

            timeoutID = setTimeout(function() {
                currentMessages = Internal.View.getOnScreenMessages();

                for (var i = 0; i < currentMessages.length; i++) {
                    msg = currentMessages[i];

                    if (lastSeenMessages.indexOf(msg) === -1) {
                        var messageID = msg.getAttribute('data-message-id');
                        Internal.Server.messageViewed(messageID);
                    }
                }

                lastSeenMessages = currentMessages;
            }, Internal.Defines.SCROLL_JITTER_TIMEOUT);
        };

        window.addEventListener('scroll', listener);
        window.addEventListener('resize', listener);
    };

    return me;
}();

Internal.Helper = function() {
    var me = {};

    /**
     * The index of node relative to parent node.
     *
     * @param {Node} node
     * @return {number}
     */
    me.index = function(node) {
        var ret = 0;
        while ((node = node.previousSibling)) {
            ret++;
        }
        return ret;
    };

    /**
     * Returns the message list element.
     *
     * @return {Node}
     */
    me.getMessageList = function() {
        var list = document.querySelector('#message-list');

        if (list == null) {
            console.error('Couldnt find message list container!');
            return null;
        }

        return list;
    };

    /**
     * Adds a bunch of test data to the message list.
     *
     * @param {number} opt_startID number to start the fake message id count
     */
    me.runDemo = function(opt_startID) {
        if (opt_startID == null) {
            opt_startID = 0;
        }

        MessageView.setMessage(JSON.stringify({ id: '' + opt_startID,
            senderID: 'senderid' + opt_startID,
            senderName: 'Jennifer Holton ' + opt_startID,
            senderImage: 'img/test/person1.png',
            body: 'Main differentiators business model micro economics ' +
            'marketplace equity <b>augmented reality</b> human computer ' +
            'interaction. <br /><br /><a href="http://google.com">' +
            'Board members super angel preferred stock.</a>' +
            '<script>alert("hacked, sucker"); console.warn("HAX");</script>',
            sentAt: 'Yesterday, 2 PM',
            attachments: [
                { id: '1', type: 'note', title: 'attachment one' },
                { id: '2', type: 'note', title: 'attachment two' }
        ]}));

        MessageView.setMessage(JSON.stringify({ id: '' + (opt_startID + 1),
            senderID: 'senderid' + (opt_startID + 1),
            senderName: 'John Marvel ' + (opt_startID + 1),
            senderImage: 'img/test/person2.png',
            body: 'Accelerator photo sharing business school drop ' +
            'out ramen hustle ' +
            'crush it revenue traction platforms. Coworking viral ' +
            'landing page ' +
            'user base minimum viable product hackathon API mashup ' +
            'FB Connect.',
            sentAt: '7 PM',
            attachments: [
        ]}));

        MessageView.setMessage(JSON.stringify({ id: '' + (opt_startID + 2),
            senderID: 'senderid' + (opt_startID + 2),
            senderName: 'Michelle Remington ' + (opt_startID + 2),
            body: 'Main differentiators business model micro economics ' +
            'marketplace equity augmented reality human computer interaction. ' +
            'Board members super angel preferred stock.',
            sentAt: '6/8/14',
            attachments: [
                { id: '1', type: 'note', title: 'attachment one' },
                { id: '2', type: 'notebook', title: 'attachment two' },
                { id: '3', type: 'notebook', title: 'attachment two' },
                { id: '4', type: 'notebook', title: 'attachment three' },
                { id: '5', type: 'note', title: 'attachment four' },
                { id: '6', type: 'notebook', title: 'attachment five' }
        ]}));

        MessageView.setMessage(JSON.stringify({ id: '' + (opt_startID + 3),
            senderID: 'senderid' + (opt_startID + 3),
            senderName: 'Ianto Jones ' + (opt_startID + 3),
            senderImage: 'img/test/ianto.png',
            body: 'There\'s stuff we don\'t know about. ' +
            'That\'s how Jack likes it.',
            sentAt: '7 PM',
            attachments: [
                { id: 'x', type: 'note', title: 'Secrets' },
                { id: 'y', type: 'notebook', title:
                    'Book of Secrets, interesting in girth as it is in length' },
        ]}));

    };

    /**
     * A longer demo script that adds a whole bunch of messages to the list.
     */
    me.runLongDemo = function() {
        for (var i = 0; i < 20; i++) {
            me.runDemo(i * 4);
        }
    };

    /**
     * Returns the name of the placeholder image that should be used when
     * a user's avatar can't be loaded / isn't specified.
     *
     * @return {string}
     */
    me.getPlaceholderImageName = function() {
        return (Internal.Runtime.isMacPlatform() ?
                'img/mac-person-placeholder.png' :
                'img/win-person-placeholder.png');
    };

    return me;
}();

Internal.Template = function() {
    var me = {};

    /**
     * Returns a document fragment representing the specified template.
     *
     * @param {string} templateName
     * @return {DocumentFragment}
     */
    me.getTemplate = function(templateName) {
        var template = document.querySelector('#' + templateName);
        var frag = document.createDocumentFragment();
        var tmpDiv = document.createElement('div');

        if (template == null) {
            return null;
        }

        // FIXME TAKE BETTER CARE AROUND SCRIPT INJECTION
        tmpDiv.innerHTML = template.innerHTML;

        while (tmpDiv.hasChildNodes()) {
            // We only accept elements in templates. Stuff like text nodes
            // are filtered.
            if (tmpDiv.firstChild.nodeType !== Node.ELEMENT_NODE) {
                tmpDiv.removeChild(tmpDiv.firstChild);
            } else {
                frag.appendChild(tmpDiv.firstChild);
            }
        }

        return frag;
    };

    /**
     * Extracts the attachment template from the DOM and returns a version
     * that's ready for editing.
     *
     * @param {!Internal.Attachment} attachment
     * @return {DocumentFragment}
     */
    me.makeAttachment = function(attachment) {
        /** @type {DocumentFragment} */
        var template = me.getTemplate('attachment-template');
        /** @type {Node} */
        var mainDiv = null;

        // TODO:athorp:2014-08-12 enforce that type must be {note,notebook}
        if (template == null ||
            attachment == null ||
            attachment.id == null ||
            attachment.type == null) {

            console.error('Couldnt make attachment template!');
            return null;
        }

        attachment.title = attachment.title || 'Untitled';
        mainDiv = template.querySelector('.attachment');

        mainDiv.setAttribute('data-attachment-id', attachment.id);
        mainDiv.classList.add(attachment.type);
        mainDiv.textContent = attachment.title;

        mainDiv.onclick = function() {
            Internal.Server.openAttachment(attachment.id);
        };

        return template;
    };

    /**
     * Attempts to safely load the given string into the given element, while
     * avoiding script injection and respecting the whitelist of elements
     * that we support, and their attributes.
     *
     * @param {Node} node
     * @param {string} textContent
     */
    var safeLoadContent = function(node, textContent) {
        /** @type {DocumentFragment} */
        var frag = document.createDocumentFragment();
        /** @type {Node} */
        var root = document.createElement('div');
        /** @type {NodeList} */
        var children = null;
        /** @type {Node} */
        var child = null;
        /** @type {boolean} */
        var success = true;
        /** @type {Array.<string>} */
        var validTags = Object.keys(Internal.Defines.STYLE_WHITELIST);
        /** @type {Array.<string>} */
        var validAttrs = null;
        /** @type {string} */
        var tag = '';
        /** @type {Text} */
        var newTextNode= null;

        root.innerHTML = textContent;
        frag.appendChild(root);
        children = root.querySelectorAll('*');

        // TODO split each case into a local closure function
        for (var i = children.length - 1; i >= 0; i--) {
            child = children[i];

            if (child.nodeType === Internal.Defines.NODE_ELEMENT_NODE) {
                // Major case: check node tag & attributes
                tag = child.tagName.toLowerCase();

                // We have a valid tag
                if (validTags.indexOf(tag) !== -1) {
                    validAttrs = Internal.Defines.NODE_ELEMENT_NODE[tag];

                    try {
                        for (var j = 0; j < child.attributes.length; j++) {
                            child.removeAttribute(child.attributes[j]);
                        }
                    } catch(err) {
                        console.warn('Failed to remove attribute from:', child);
                        success = false;
                        break;
                    }
                } else {
                    try {
                        newTextNode = document.createTextNode(child.outerHTML);
                        child.parentNode.replaceChild(newTextNode, child);
                    } catch(err) {
                        console.warn('Failed to replace with text:', child);
                        success = false;
                        break;
                    }
                }

            } else if (child.nodeType === Internal.Defines.NODE_TEXT_NODE) {
                // Text nodes are safe / fine
                continue;

            } else {
                // Means that we found some 'exotic' node that we don't need
                try {
                    child.parentNode.removeChild(child);
                } catch(err) {
                    console.warn('Failed to remove child', child);
                    success = false;
                    break;
                }
            }
        }

        if (success === false) {
            console.warn('Failed to safe load content:', frag);
            node.appendChild(document.createTextNode('Malformed Message!'));
            return;
        }

        // Copy everything back into the given node
        while (root.firstChild) {
            node.appendChild(root.firstChild);
        }
    };

    /**
     * Extracts the message template from the DOM and returns a version
     * that's ready for editing.
     *
     * @param {!Internal.Message} message
     * @return {DocumentFragment}
     */
    me.makeMessage = function(message) {
        /** @type {DocumentFragment} */
        var template = me.getTemplate('message-template');

        if (template == null ||
            message == null ||
            message.id == null ||
            message.senderID == null) {

            console.error('Couldnt make message template!');
            return null;
        }

        message.senderName = message.senderName || '';
        message.senderImage = message.senderImage ||
            Internal.Helper.getPlaceholderImageName();
        message.body = message.body || '';
        message.sentAt = message.sentAt || '';
        message.attachments = message.attachments || [];

        /** @type {Node} */ var main = template.querySelector('.message');
        /** @type {Node} */ var image = template.querySelector('.senderImage');
        /** @type {Node} */ var name = template.querySelector('.senderName');
        /** @type {Node} */ var body = template.querySelector('.body');
        /** @type {Node} */ var sentAt = template.querySelector('.sentAt');
        /** @type {Node} */ var attachments = template.querySelector('.attachments');

        main.setAttribute('data-message-id', message.id);
        image.setAttribute('data-desired-src', message.senderImage);
        Internal.Runtime.tryLoadImage(image);
        name.textContent = message.senderName;
        name.setAttribute('data-sender-id', message.senderID);
        safeLoadContent(body, message.body);
        sentAt.textContent = message.sentAt;

        for (var i = 0; i < message.attachments.length; i++) {
            attachments.appendChild(me.makeAttachment(message.attachments[i]));
        }

        return template;
    };

    return me;
}();

Internal.View = function() {
    var me = {};

    /**
     * Scrolls to the bottom of the message list.
     */
    me.scrollToBottom = function() {
        // FIXME believe it or not, document.body.scrollHeight is a major
        // bottleneck when I profiled this code by loading a massive amount of
        // messages at once. I get a 10x speedup when I remove it here.
        // I should de-jitter this function when it's called repeatedly.
        window.scrollTo(0, document.body.scrollHeight);
    };

    /**
     * Returns a reference to the message DOM element if it's already
     * been rendered. If it hasn't already been rendered, then this returns null.
     *
     * @param {string} messageID
     * @return {Node|null}
     */
    me.getMessageDOMByID = function(messageID) {
        /** @type {Node} */ var messageList = Internal.Helper.getMessageList();
        /** @type {Node} */ var childNode;

        for (var i = messageList.childNodes.length - 1; i >= 0; i--) {
            childNode = messageList.childNodes[i];

            if (childNode == null ||
                    childNode.classList == null ||
                    childNode.classList.contains('message') === false) {
                console.warn('Message list contains weird node,', childNode);
                continue;
            }

            if (childNode.getAttribute('data-message-id') === messageID) {
                return childNode;
            }
        }

        return null;
    };

    /**
     * Returns whether the given message element is currently scrolled into view.
     *
     * @param {Node} message
     * @return {boolean}
     */
    me.isOnScreen = function(message) {
        var scrollTop = window.scrollY;
        var scrollBottom = scrollTop + window.innerHeight;

        // Normalized to be relative to the top of window, rather than viewport
        var elementTop = message.getBoundingClientRect().top + scrollTop;
        var elementBottom = message.getBoundingClientRect().bottom + scrollTop;

        // Generate "common interval" between the two "lines"
        var commonTop = Math.max(scrollTop, elementTop);
        var commonBottom = Math.min(scrollBottom, elementBottom);

        // Percentage that the message is on screen
        var overlapPercent = Math.max(0,
                Math.round((commonBottom - commonTop) /
                    (elementBottom - elementTop) * 100));

        return (overlapPercent >= Internal.Defines.MESSAGE_ON_SCREEN_PERCENT);

    };

    /**
     * Returns a list of messageIDs for messages that are currently on screen.
     *
     * @return {Array.<string>}
     */
    me.getOnScreenMessages = function() {
        var messages = Internal.Helper.getMessageList().childNodes;
        var onScreen = [];

        for (var i = 0; i < messages.length; i++) {
            if (me.isOnScreen(messages[i])) {
                onScreen.push(messages[i]);
            }
        }

        return onScreen;
    };

    /**
     * Adds a message to the message list.
     *
     * @param {!(Node|DocumentFragment)} message
     */
    me.addMessage = function(message) {
        Internal.Helper.getMessageList().appendChild(message);
    };

    /**
     * Scrolls the view to show the specified message.
     *
     * @param {string} messageID
     * @param {boolean} opt_bottom
     */
    me.scrollToMessage = function(messageID, opt_bottom) {
        /** @type {Node} */ var message = me.getMessageDOMByID(messageID);

        if (opt_bottom == null) {
            opt_bottom = false;
        }

        if (message == null) {
            console.error('No message with ID', messageID);
            return;
        }

        message.scrollIntoView(!opt_bottom);
    };

    /**
     * Wipes entire display, removing everything.
     */
    me.clear = function() {
        var list = Internal.Helper.getMessageList();

        // A while loop is considerably faster than innerHTML = ''
        while (list.firstChild) {
            list.removeChild(list.firstChild);
        }
    };

    return me;
}();

Internal.Controller = function() {
    var me = {};

    /**
     * Creates and inserts a message given a Message object.
     *
     * @param {string} message
     */
    me.setMessage = function(message) {
        var decode = /** @type {Internal.Message} */ (JSON.parse(message));
        var messageDOM = Internal.View.getMessageDOMByID(decode.id);
        var frag = Internal.Template.makeMessage(decode);

        if (messageDOM != null) {
            if (messageDOM.parentNode) {
                messageDOM.parentNode.replaceChild(frag, messageDOM);
            } else {
                console.warn('Couldnt replace message!', messageDOM, frag);
            }
        } else {
            Internal.View.addMessage(frag);
            Internal.View.scrollToBottom();
        }
    };

    return me;
}();

// Exports
MessageView.setMessage = Internal.Controller.setMessage;
MessageView.clear = Internal.View.clear;
MessageView.scrollToMessage = Internal.View.scrollToMessage;

// Startup
Internal.Runtime.loadCSS();
Internal.Runtime.imageLoadLoop();
Internal.Runtime.messagesOnScreenWatcher();
